"""Holds config utilities"""
import os
import sys
from typing import Dict, List, Optional, Tuple
import networkx

import hcl2
import jsons
import yaml
from jsons import DeserializationError
from ssm_cache import SSMParameterGroup

from terrawrap.exceptions import NotTerraformConfigDirectory, NoDependency
from terrawrap.models.wrapper_config import (
    WrapperConfig,
    AbstractEnvVarConfig,
    SSMEnvVarConfig,
    TextEnvVarConfig,
    BackendsConfig,
    UnsetEnvVarConfig,
)
from terrawrap.utils.collection_utils import update
from terrawrap.utils.path import get_absolute_path, calc_repo_path

DEFAULT_REGION = 'us-west-2'
SSM_ENVVAR_CACHE = SSMParameterGroup(max_age=600)
TF_WRAP_FILE = ".tf_wrapper"


def find_variable_files(path: str) -> List[str]:
    """
    Convenience function for finding all Terraform variable files by walking a given path.
    :param path: The path to the Terraform configuration directory.
    :return: A list of Terraform variable files that can be found by walking down that path, in order of
    discovery from the root of the path.
    """
    variable_files = []

    elements = path.split(os.path.sep)

    cur_path = os.path.sep

    for element in elements:
        cur_path = os.path.join(cur_path, element)
        for file in os.listdir(cur_path):
            if file.endswith(".auto.tfvars"):
                variable_files.append(os.path.join(cur_path, file))

    return variable_files


def find_wrapper_config_files(path: str) -> List[str]:
    """
    Convenience function for finding all wrapper config files by walking a given path.
    :param path: The path to the Terraform configuration directory.
    :return: A list of wrapper config files that can be found by walking down that path, in order of
    discovery from the root of the path.
    """
    wrapper_config_files = []

    elements = path.split(os.path.sep)

    cur_path = os.path.sep

    for element in elements:
        cur_path = os.path.join(cur_path, element)
        for file in os.listdir(cur_path):
            if file.endswith(TF_WRAP_FILE):
                wrapper_config_files.append(os.path.join(cur_path, file))

    return wrapper_config_files


def parse_wrapper_configs(wrapper_config_files: List[str]) -> WrapperConfig:
    """
    Function for parsing the Terraform wrapper config file.
    :param wrapper_config_files: A list of file paths to wrapper config files. Config files later in the list
    override those earlier in the list, and are merged with the default config and earlier files.
    :return: A WrapperConfig object representing the accumulated values of all the wrapper config files
    """
    generated_wrapper_config: Dict = {}

    for wrapper_config_path in wrapper_config_files:
        with open(wrapper_config_path, encoding='utf-8') as wrapper_config_file:
            wrapper_config = yaml.safe_load(wrapper_config_file)
            if wrapper_config and isinstance(wrapper_config, dict):
                generated_wrapper_config = update(generated_wrapper_config, wrapper_config)

    try:
        wrapper_config_obj: WrapperConfig = jsons.load(generated_wrapper_config, WrapperConfig, strict=True)
        return wrapper_config_obj
    except DeserializationError as exception:
        print(f"Cannot parse wrapper config from files: {wrapper_config_files}")
        raise exception


def is_config_directory(directory: str) -> bool:
    """
    Checks if a wrapper file directory is a config_directory
    :param directory: A wrapper files directory
    :return: A boolean True if the config section of the wrapper file is not False or doesn't exist.
    """
    config = False
    for file in os.listdir(directory):
        if file.endswith(".tf"):
            config = True
    return config


def create_wrapper_config_obj(config_dir, wrapper_file=None):
    """
    Given a config dir containing a tf_wrapper.
    Parses the tf_wrapper for dependencies and makes a a wrapper config object.
    :param config_dir: A tf directory containing a tf_wrapper.
    :param wrapper_file: A wrapper file passed in if known
    :return: wrapper_config_obj: a wrapper config object
    """
    if not wrapper_file:
        for file in os.listdir(config_dir):
            if file.endswith(TF_WRAP_FILE):
                wrapper_file = os.path.join(config_dir, file)

    wrapper_files = [wrapper_file] if wrapper_file else []
    wrapper_config_obj: WrapperConfig = parse_wrapper_configs(wrapper_files)
    if wrapper_config_obj.depends_on:
        depends_on = []
        for dependency in wrapper_config_obj.depends_on:
            abs_dependency = get_absolute_path(dependency)
            if not os.path.isdir(abs_dependency):
                abs_dependency = get_absolute_path(dependency, config_dir)
            depends_on.append(abs_dependency)
        wrapper_config_obj.depends_on = depends_on
    if not is_config_directory(config_dir):
        wrapper_config_obj.config = False
    return wrapper_config_obj


def walk_and_graph_directory(starting_dir: str, config_dict) -> Tuple[networkx.DiGraph, List[str]]:
    """
    Given a starting directory, walks it and returns all dependency info.
    :param starting_dir: The starting directory
    :param config_dict: A dictionary containing wrapper config objects for each seen directory
    :return: directory_graph: A graph composed of all dependency information for a directory
    """
    graph_list = []
    post_graph_runs = []
    for root, _, files in os.walk(starting_dir):
        has_tf_wrapper = False
        for file in files:
            if file.endswith(TF_WRAP_FILE):
                has_tf_wrapper = True
                wrapper_file = os.path.join(root, file)
                wrapper_config_obj = create_wrapper_config_obj(root, wrapper_file)
                if not wrapper_config_obj.config:
                    continue
                if not wrapper_config_obj.apply_automatically:
                    continue
                if wrapper_config_obj.depends_on is None:
                    post_graph_runs.append(root)
                    continue
                single_config_dependency_graph = networkx.DiGraph()
                visited: List[str] = []
                graph_wrapper_dependencies(root, config_dict, single_config_dependency_graph, visited)
                graph_list.append(single_config_dependency_graph)
        if not has_tf_wrapper and is_config_directory(root):
            post_graph_runs.append(root)
    directory_graph = networkx.compose_all(graph_list)

    return directory_graph, post_graph_runs


def walk_without_graph_directory(starting_dir: str) -> List[str]:
    """
    Given a starting directory, walks it and returns a list of all tf configs to apply.
    :param starting_dir: The starting directory
    :return: post_graph: A graph composed of all dependency information for a directory
    """
    post_graph_runs = []
    for root, _, files in os.walk(starting_dir):
        has_tf_wrapper = False
        for file in files:
            if file.endswith(TF_WRAP_FILE):
                has_tf_wrapper = True
                wrapper_file = os.path.join(root, file)
                wrapper_config_obj = create_wrapper_config_obj(root, wrapper_file)
                if wrapper_config_obj.depends_on is not None:
                    raise NoDependency("Discovered dependency information")
                if not wrapper_config_obj.config:
                    continue
                if not wrapper_config_obj.apply_automatically:
                    continue
                post_graph_runs.append(root)
        if not has_tf_wrapper and is_config_directory(root):
            post_graph_runs.append(root)

    return post_graph_runs


# pylint: disable=R0912
def graph_wrapper_dependencies(config_dir: str, config_dict, graph: networkx.DiGraph, visited: List[str]):
    """
    Given a directory, recursively finds all other directories it depends on and builds a graph.
    :param config_dir: The config directory to obtain a dependency graph for
    :param config_dict: A dictionary containing wrapper config objects for each seen directory
    :param graph: The graph to add dependency info to. Empty at first, reused in recursion.
    :param visited: A list of visited nodes. Empty at first, reused in recursion
    """
    if config_dir in visited:
        return
    visited.append(config_dir)

    if config_dict.get(config_dir):  # add to dictionary so we only read the file once
        wrapper_config_obj = config_dict[config_dir].get("wrapper_config")
    else:
        wrapper_config_obj = create_wrapper_config_obj(config_dir)
        config_dict[config_dir] = {
            "wrapper_config": wrapper_config_obj
        }

    if wrapper_config_obj.config:
        graph.add_node(config_dir)

    tf_dependencies = wrapper_config_obj.depends_on

    if tf_dependencies is None:
        print("Cannot list a dependency without tf_wrapper dependency configuration:", config_dir)
        sys.exit(1)

    for dependency in tf_dependencies:
        graph.add_node(dependency)
        if config_dir in graph:
            graph.add_edge(dependency, config_dir)

    wrappers = find_wrapper_config_files(config_dir)
    wrappers.reverse()  # we want the closest wrapper file that gives inherited dependencies
    for wrapper in wrappers:
        wrapper_dir = os.path.dirname(wrapper)
        if config_dict.get(wrapper_dir):  # add to dictionary so we only read the file once
            new_wrapper_config_obj = config_dict[wrapper_dir].get("wrapper_config")
        else:
            new_wrapper_config_obj = create_wrapper_config_obj(wrapper_dir)
            config_dict[wrapper_dir] = {
                "wrapper_config": new_wrapper_config_obj
            }

        if wrapper_dir == config_dir:
            continue
        if new_wrapper_config_obj.depends_on is not None:
            inherited_dependencies = new_wrapper_config_obj.depends_on
            added = False
            for dependency in inherited_dependencies:
                if dependency == config_dir:
                    continue
                added = True
                graph.add_node(dependency)
                if config_dir in graph:
                    graph.add_edge(dependency, config_dir)
            if added:
                break  # we only need the closest, the recursion will handle anything higher

    for predecessor in list(graph.predecessors(config_dir)):
        graph_wrapper_dependencies(predecessor, config_dict, graph, visited)


def resolve_envvars(envvar_configs: Dict[str, AbstractEnvVarConfig]) -> Dict[str, str]:
    """
    Resolves the 'envvars' section from the wrapper config to actual environment variables that can be easily
    supplied to a command.
    :param envvar_configs: The 'envvars' dictionary from the wrapper config.
    :return: A dictionary representing the environment variables that were resolved, with the key being the
    name of the environment variable and the value being the value of the environment variable.
    """
    resolved_envvars = {}
    for envvar_name, envvar_config in envvar_configs.items():
        if isinstance(envvar_config, SSMEnvVarConfig):
            resolved_envvars[envvar_name] = SSM_ENVVAR_CACHE.parameter(envvar_config.path).value
        if isinstance(envvar_config, TextEnvVarConfig):
            resolved_envvars[envvar_name] = str(envvar_config.value)
        if isinstance(envvar_config, UnsetEnvVarConfig):
            resolved_envvars[envvar_name] = None
    return resolved_envvars


def calc_backend_config(
        path: str,
        variables: Dict[str, str],
        wrapper_config: WrapperConfig,
        existing_backend_config: BackendsConfig
) -> List[str]:
    """
    Convenience function for calculating the backend config of the given Terraform directory.
    :param path: The path to the directory containing the Terraform config.
    :param variables: The variables derived from the auto.tfvars files.
    :param wrapper_config:
    :param existing_backend_config: Backend config object from parsing the terraform resource in the configs
    :return: A dictionary representing the backend configuration for the Terraform directory.
    """

    backend_config = ['-reconfigure']
    options: Dict[str, str] = {}
    repo_path = calc_repo_path(path=path)

    # for backwards compatibility, include the default s3 backend options we used to automatically include
    if existing_backend_config.s3 is not None:
        region = variables.get('region', '')
        account_short_name = variables.get('account_short_name')
        terraform_bucket = f"{region}--mclass--terraform--{account_short_name}"

        options = {
            'dynamodb_table': variables.get('terraform_lock_table', 'terraform-locking'),
            'encrypt': 'true',
            'key': f'{repo_path}.tfstate',
            'region': region,
            'bucket': variables.get('terraform_state_bucket', terraform_bucket),
            'skip_region_validation': 'true',
            'skip_credentials_validation': 'true'
        }

    # copy any backend options from the backend config
    if wrapper_config.backends:
        wrapper_options: Dict[str, Optional[str]] = {}
        if existing_backend_config.gcs is not None and wrapper_config.backends.gcs is not None:
            # convert the object into a dict so we can append each field to the backend config dynamically
            wrapper_options = vars(wrapper_config.backends.gcs)
            wrapper_options['prefix'] = repo_path
        if existing_backend_config.s3 is not None and wrapper_config.backends.s3 is not None:
            wrapper_options = vars(wrapper_config.backends.s3)
        options.update({key: value for key, value in wrapper_options.items() if value is not None})

    backend_config.extend([f'-backend-config={key}={value}' for key, value in options.items()])
    return backend_config


def parse_variable_files(variable_files: List[str]) -> Dict[str, str]:
    """
    Convenience function for parsing variable files and returning the variables as a dictionary
    :param variable_files: List of file paths to variable files. Variable files overwrite files before them.
    :return: A dictionary representing the contents of those variable files.
    """
    variables: Dict = {}

    for variable_file in variable_files:
        with open(variable_file, encoding='utf-8') as var_file:
            variables.update(hcl2.load(var_file).items())

    return variables


def parse_backend_config_for_dir(dir_path: str) -> Optional[BackendsConfig]:
    """
    Parse tf files in a directory and try to get the state backend config from the "terraform" resource
    :param dir_path: Directory that has tf files
    :return: Backend config if a "terraform" resource exists, otherwise None
    """
    has_tf_files = False
    for file_path in os.listdir(dir_path):
        if '.terraform' in file_path or not file_path.endswith('.tf'):
            continue

        has_tf_files = True

        result = _parse_backend_config_for_file(
            file_path=os.path.join(dir_path, file_path),
        )
        if result:
            return result

    if not has_tf_files:
        raise NotTerraformConfigDirectory()

    return None


def _parse_backend_config_for_file(file_path: str) -> Optional[BackendsConfig]:
    with open(file_path, encoding='utf-8') as tf_file:
        try:
            configs: Dict[str, List] = hcl2.load(tf_file)

            terraform_config_blocks: List[Dict] = configs.get('terraform', [])
            for terraform_config in terraform_config_blocks:
                if 'backend' in terraform_config:
                    return jsons.load(terraform_config['backend'][0], BackendsConfig, strict=True)
            return None
        except Exception:
            print(f'Error while parsing file {file_path}')
            raise



""" Holds graph utilities"""

import os
from pathlib import Path
from typing import List, Tuple, Any, Dict, Set
import networkx
from terrawrap.exceptions import NoDependency
from terrawrap.utils.config import walk_without_graph_directory


def has_cycle(graph: networkx.DiGraph) -> bool:
    """
    Checks that a graph does not contain a cycle.
    :param graph: The graph to check
    :return: A boolean true if there is a cycle.
    """
    sources = find_source_nodes(graph)
    if not sources:
        return True

    for source in sources:
        try:
            cycle = networkx.find_cycle(graph, source)
            if cycle:
                print(cycle)
                return True
        except networkx.NetworkXNoCycle:
            return False
    return False


def find_source_nodes(graph: networkx.DiGraph) -> List[str]:
    """
    For a given graph return a list of source nodes (Nodes with no predecessors)
    :param graph: The graph to look for source nodes in.
    :return: source_nodes -  a list of nodes (str) in the graph with no predecessors
    """
    source_nodes = []
    for node in graph:
        preds = list(graph.predecessors(node))
        if preds:
            continue
        source_nodes.append(node)

    return source_nodes


def successors(depth: int, node: str, graph: networkx.DiGraph) -> Tuple[int, str, List[str]]:
    """
    For a given node in a given graph, return the node with it's successors.
    :param depth: The current depth of the successors
    :param node: The node to obtain successors for
    :param graph: The graph to query for successors
    :return: successors -  a tuple containing a node and a list of its successors
    """
    node_successors = (depth, node, list(graph.successors(node)))
    return node_successors


def generate_dependencies(nodes: List[str], graph: networkx.DiGraph) -> List[Any]:
    """
    Creates a list of dependencies for a graph which
    contains lists of tuples with a node its depth and its successors
    :param nodes: A list of nodes
    :param graph: A graph containing dependency information
    :return: A list of all dependency information in the graph
    """
    dependencies = []
    depth = 1
    for node in nodes:
        path = []
        node_successors = successors(depth, node, graph)
        path.append(node_successors)
        if node_successors[1]:
            generate_helper(node_successors[2], graph, depth + 1, path)
        dependencies.append(path)
    return dependencies


def generate_helper(nodes: List[str], graph: networkx.DiGraph, depth: int, path: List[Any]) -> List[str]:
    """
    The recursive helper function for generate_dependencies
    :param nodes: A list of nodes
    :param graph: The graph to search
    :param depth: The current depth
    :param path: The current path the dependencies are on
    :return: THe updated path the dependencies are in.
    """
    for node in nodes:
        node_successors = successors(depth, node, graph)
        path.append(node_successors)
        if node_successors[1]:
            generate_helper(node_successors[2], graph, depth + 1, path)
    return path


def visualize(dependencies: List[List[str]]):
    """
    For a given graph, generate a human readable output
    :param dependencies: A list of paths for a graph
    """
    for path in dependencies:
        depth = 0
        for node in path:
            tab_spacing = int(node[0]) - 1 + depth
            if depth > 0:
                print()
            relative_node = node[1].replace(os.getcwd(), "")
            print(("\t" * tab_spacing) + ">", relative_node)
        depth += 1


def find_symlink_directories(graph: networkx.DiGraph) -> List[Path]:
    """
    Finds all symlink directories in a given graph
    :param graph: The graph to find symlinks in
    :return: symlinks: A list of directories that are symlinks
    """
    symlinks = []

    for node in graph:
        path = Path(node)
        if path.is_symlink():
            symlinks.append(path)

    return symlinks


def connect_symlinks(graph: networkx.DiGraph, symlink_dict: Dict[str, Set[str]]):
    """
    Implements dependency linking in the graph for chain-linked directories.
    This is necessary because we must run symlinked directories serially in respect to each other
    :param graph: The graph containing symlinked directories
    :param symlink_dict: A dictionary of symlinks and the directories that are linked to them.
    """
    for symlink in symlink_dict.keys():
        current = symlink
        secondary = 0
        links = list(symlink_dict[symlink])
        while secondary != len(links):
            secondary_link = links[secondary]

            # TODO: implement better way for symlinked tf configs with terraform config
            try:
                links_to_run = walk_without_graph_directory(secondary_link)
                for link in links_to_run:
                    graph.add_edge(current, link)
            except NoDependency as error:
                print(error)
                graph.add_edge(current, secondary_link)

            current = secondary_link
            secondary += 1



"""Contains functions for checking the latest version of this package"""
import sys
import tempfile
import os
from time import sleep

import requests
from packaging import version
from diskcache import Cache


ONE_DAY_IN_SECONDS = 60 * 60 * 24
cache = Cache(os.path.join(tempfile.gettempdir(), "terrawrap_version_cache"))


def version_check(current_version: str) -> bool:
    """
    Print a warning message if a stale version of Terrawrap is detected.
    :param current_version: The currently installed version of Terrawrap.
    :return: True if the version of Terrawrap is stale.
    """
    try:
        latest_version = get_latest_version(current_version=current_version)
        if version.parse(latest_version) <= version.parse(current_version):
            return False

        print(
            "WARNING: Your version of Terrawrap is stale!",
            f"You have version '{current_version}' but the latest is '{latest_version}'",
            "Please upgrade as soon as possible!\n pip install --upgrade terrawrap \n",
            sep="\n",
            file=sys.stderr,
        )
        sleep(1)
        return True
    except Exception as exp:
        print(
            f"WARNING: Encountered some error while checking for latest version of Terrawrap: {repr(exp)}",
        )
    return False


# We supply current_version so that the cache is invalidated when you install a new version of Terrawrap.
# pylint: disable=unused-argument
@cache.memoize(expire=ONE_DAY_IN_SECONDS)
def get_latest_version(current_version: str) -> str:
    """
    Get the latest version of Terrawrap from Pypi. Caches this lookup for one day locally.
    :param current_version: The current version of Terrawrap.
    :return: The latest version of Terrawrap, potentially delayed by one day.
    """
    response = requests.get("https://pypi.python.org/pypi/terrawrap/json", timeout=5).json()
    return response["info"]["version"]



"""Module for interacting with DynamoDB"""
from typing import Any, Dict

import boto3
from amplify_aws_utils.resource_helper import throttled_call


class DynamoDB:
    """Class for operating with DynamoDB"""

    def __init__(self, region: str, client=None):
        self.client = client or boto3.client('dynamodb', region_name=region)

    def upsert_item(
            self,
            table_name: str,
            primary_key_name: str,
            primary_key_value: str,
            attribute_name: str,
            attribute_value: str
    ) -> Dict[str, Any]:
        """
        Insert/update items in a DynamoDB table for a specific primary key value.
        The value inserted/updated is of type string.
        :param table_name: DynamoDB table name
        :param primary_key_name: Table's primary key name
        :param primary_key_value: Table's primary key value
        :param attribute_name: Item attribute to be inserted/updated
        :param attribute_value: Attribute value to be inserted/updated
        :return: response
        """
        key = {primary_key_name: {'S': primary_key_value}}
        expression_attribute_values = {':d': {'S': attribute_value}}
        update_expression = f'SET {attribute_name} = :d'

        return throttled_call(
            self.client.update_item,
            TableName=table_name,
            Key=key,
            ExpressionAttributeValues=expression_attribute_values,
            UpdateExpression=update_expression
        )

    def delete_item(
            self,
            table_name: str,
            primary_key_name: str,
            primary_key_value: str,
    ):
        """
        Convenience function for deleting an item from a DynamoDB table with a specific primary key value.
        :param table_name: DynamoDB table name
        :param primary_key_name: Table's primary key name
        :param primary_key_value: Table's primary key value
        :return: response
        """
        key = {primary_key_name: {'S': primary_key_value}}

        return throttled_call(
            self.client.delete_item,
            TableName=table_name,
            Key=key,
        )



"""Module for containing CLI convenience functions"""
from __future__ import print_function

import logging
import subprocess
import tempfile
import time
from enum import Enum

from typing import List, Tuple, Union

import requests

from amplify_aws_utils.resource_helper import Jitter
from aws_requests_auth.boto_utils import BotoAWSRequestsAuth

from terrawrap.utils.git_utils import get_git_root

logger = logging.getLogger(__name__)

MAX_RETRIES = 5
RETRIABLE_ERRORS = [
    'RequestError: send request failed',
    'unexpected EOF',
    'Throttling',
    'timeout while waiting for state',
    'ServiceUnavailable: Service Unavailable',
    'failed to decode query XML error response',
    'connection reset',
    'Connection reset',
    'Please try again.',
    'Client.Timeout exceeded',
    'Request limit for operation',
    'try again later',
    'handshake timeout',
    'SSL_ERROR_SYSCALL',
    'Api Rate Limit Exceeded',
    'TooManyUpdates',
]
AUDIT_POST_PATH = '/audit_info'
AUDIT_UPDATE_PATH = '/update_audit_info'


class Status(str, Enum):
    """Enum for status of execute_command"""
    SUCCESS = 'SUCCESS'
    IN_PROGRESS = 'IN PROGRESS'
    FAILED = 'FAILED'


def execute_command(
        args: Union[List[str], str],
        *pargs,
        print_output: bool = True,
        capture_stderr: bool = True,
        print_command: bool = False,
        retry: bool = False,
        timeout: int = 15 * 60,
        audit_api_url: str = None,
        **kwargs
) -> Tuple[int, List[str]]:
    """
    Convenience function for executing a given command and optionally printing the output.
    :param args: List of arguments to execute, or a single string.
    :param pargs: Any additional positional arguments to Popen.
    :param print_output: True if the output of the command should be printed immediately. Defaults to True.
    :param capture_stderr: True if stderr should be captured. Defaults to True.
    :param print_command: True if the command should be printed before executing. Defaults to False.
    :param timeout: Max amount of time to keep retrying to execute command. Defaults to 15 minutes.
    :param retry: Retry a number of times if network errors. Defaults to False.
    :param audit_api_url: Audit API URL to submit POST request. Defaults to None so no data is sent.
    :param kwargs: Any additional keyword arguments to Popen.
    :return: A tuple of the exit code and output of the command.
    """
    try_count = 0

    # It's possible for an envvar to be set to none, so exclude those envvars.
    if 'env' in kwargs:
        kwargs['env'] = {
            key: value
            for key, value in kwargs['env'].items()
            if value is not None
        }

    # Get time - nanoseconds since epoch
    start_time = int(time.time())

    if audit_api_url and kwargs['cwd'] and 'apply' in args:
        # Call _post_audit_info for working directory, setting status to 'in progress'
        _post_audit_info(
            audit_api_url=audit_api_url,
            path=kwargs['cwd'],
            start_time=start_time,
        )
    else:
        logger.info('No audit_api_url provided')

    jitter = Jitter()
    time_passed = 0
    exit_code = 0
    stdout: List[str] = []
    while try_count < MAX_RETRIES if retry else 1:
        exit_code, stdout = _execute_command(
            args,
            print_output,
            capture_stderr,
            print_command,
            *pargs,
            **kwargs,
        )

        try_count += 1

        network_errors = _get_retriable_errors(stdout)
        if exit_code != 0 and network_errors and retry:
            logger.warning('Found network errors while running %s command: %s', args, network_errors)
        else:
            # The command either succeeded or failed with a non network error. don't retry
            break

        if time_passed >= timeout:
            break

        time_passed = jitter.backoff()

    if audit_api_url and kwargs['cwd'] and 'apply' in args:
        # Call _post_audit_info again, this time to update the 'in progress' entry with new status and output
        _post_audit_info(
            audit_api_url=audit_api_url,
            path=kwargs['cwd'],
            exit_code=exit_code,
            stdout=stdout,
            start_time=start_time,
            update=True
        )
    else:
        logger.info('No audit_api_url provided')

    if time_passed >= timeout:
        raise TimeoutError(f'Timed out retrying {args} command')

    return exit_code, stdout


def _execute_command(
        args: Union[List[str], str],
        print_output: bool,
        capture_stderr: bool,
        print_command: bool,
        *pargs,
        **kwargs
) -> Tuple[int, List[str]]:
    """
    Private function for executing a given command and optionally printing the output.
    :param args: List of arguments to execute, or a single string.
    :param print_output: True if the output of the command should be printed immediately. Defaults to True.
    :param capture_stderr: True if stderr should be captured. Defaults to True.
    :param print_command: True if the command should be printed before executing. Defaults to False.
    :param pargs: Any additional positional arguments to Popen.
    :param kwargs: Any additional keyword arguments to Popen.
    :return: A tuple of the exit code and output of the command.
    """
    stdout_write, stdout_path = tempfile.mkstemp()
    with open(stdout_path, "rb") as stdout_read, open('/dev/null', 'w', encoding='utf-8') as dev_null:

        if print_command:
            print(f"Executing: {' '.join(args)}")

        kwargs['stdout'] = stdout_write
        kwargs['stderr'] = stdout_write if capture_stderr else dev_null

        # pylint: disable=consider-using-with
        process = subprocess.Popen(
            args,
            *pargs,
            **kwargs
        )

        while True:
            output = stdout_read.read(1).decode(errors="replace")

            if output == '' and process.poll() is not None:
                break

            if print_output and output:
                print(output, end="", flush=True)

        exit_code = process.poll()

        stdout_read.seek(0)
        stdout = [line.decode(errors="replace") for line in stdout_read.readlines()]

        # ignoring mypy error below because it thinks exit_code can sometimes be None
        # we know that will never be the case because the above While loop will keep looping forever
        # until exit_code is not None
        return exit_code, stdout  # type: ignore


def _get_retriable_errors(out: List[str]) -> List[str]:
    """Filter line output for retriable errors"""
    return [
        line for line in out
        if any(error in line for error in RETRIABLE_ERRORS)
    ]


def _post_audit_info(
        audit_api_url: str,
        path: str,
        start_time: int,
        exit_code: int = None,
        stdout: List[str] = None,
        update: bool = False
):
    root = get_git_root(path)
    path = path.replace(root, '')

    status = Status.IN_PROGRESS if exit_code is None else (
        Status.SUCCESS if exit_code == 0 else Status.FAILED
    )

    logger.info('Attempting to send data to Audit API: %s - %s', path, status)

    url = (audit_api_url + AUDIT_UPDATE_PATH) if update else (audit_api_url + AUDIT_POST_PATH)

    auth = BotoAWSRequestsAuth(
        aws_host='terraform-audit-api.devops.amplify.com',
        aws_region='us-west-2',
        aws_service='execute-api'
    )

    stdout_str = ''.join(stdout) if stdout else ''

    try:
        requests.post(
            url=url,
            auth=auth,
            json={
                'directory': path,
                'start_time': start_time,
                'status': status,
                'output': stdout_str
            },
            timeout=30,
        )
        logger.info('Successfully posted data to provided url: %s', audit_api_url)
    except requests.exceptions.RequestException:
        logger.error("Unable to post data to provided url: %s", audit_api_url)
